# Copyright 2022 Ashley R. Thomas
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""Derive from Singleton to create a new singleton.
"""


class Singleton(type):
    """Create a singleton by using this as a metaclass.
    Example1: class MySingleton(metaclass=Singleton):
    Example2: class MySingleton(MyBaseClass, metaclass=Singleton):
    """

    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            # Calls __init__ on new/only instance.
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        else:
            # Calls __init__ on existing/only instance (can be used to ensure same parameters or raise etc.).
            cls._instances[cls].__init__(*args, **kwargs)
        return cls._instances[cls]
